/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "chan_list.hh"
#include "checksum_crc32.hh"
#include "ParseLine.hh"
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <cctype>

static const size_t source_list_default_size=4;
static const size_t channel_list_default_size=1024;

//======================================  Make a fast swap template
namespace std {
    template<>
    inline void
    swap(sends::chan_entry::source_entry& a, 
	 sends::chan_entry::source_entry& b) {
	a.swap(b);
    }

    template<>
    inline void
    swap(sends::chan_entry& a, sends::chan_entry& b) {
	a.swap(b);
    }
}

using namespace std;
using namespace sends;

//======================================  String size for memory estimates
index_type
string_size(const std::string& str) {
    return (str.size() + 7) & (-8);
}

//======================================  Channel entry constructor
//  static const gps_type inf_gps=3999999999;

//======================================  Source entry null constructor
chan_entry::source_entry::source_entry(void)
{}

//======================================  Source entry data constructor
chan_entry::source_entry::source_entry(const std::string& name, 
				       const epoch& ep)
    : _name(name), _epoch(ep)
{}

//======================================  Source entry destructor
chan_entry::source_entry::~source_entry(void) {
}

//======================================  Source entry destructor
std::ostream& 
chan_entry::display(std::ostream& out) const {
    out << name() << " " << type().str() << " " 
	<< rate() << " " << dtype().str();
    return out;
}


//======================================  Merge source entries
void 
chan_entry::source_entry::merge(const source_entry& se) {
    _epoch.merge(se._epoch);
}

//======================================  Swap source entries
void
chan_entry::source_entry::swap(source_entry& b) {
    _name.swap(b._name);
    std::swap<epoch> (_epoch, b._epoch);
}

//======================================  Channel entry hash
chan_entry::hash_type
chan_entry::hash_key(const std::string& name) {
    checksum_crc32 crc32;
    crc32.add(name);
    return crc32.result();
}

//======================================  Channel entry constructor
chan_entry::chan_entry(const string& name, chan_type type, long access, 
		       data_type dtype, double rate, const string& source,
		       const epoch& ep) 
  : mName(name), mHash(0), mType(type), mAccess(access), mDType(dtype), 
    mRate(rate)
{
    mHash = hash_key(name);
    mSource.reserve(source_list_default_size);
    if (!source.empty()) mSource.push_back(source_entry(source, ep));
}

//======================================  Test if channel is available at time.
//
//  This is a stub function - universal availablility is cool.
bool 
chan_entry::available(gps_type t) const {
    return true;
}

//======================================  Add a source entry
void 
chan_entry::addSource(const std::string& ftype, const epoch& ep) {
    source_entry newent(ftype, ep);
    source_iter itr = mSource.lower_bound(newent);
    if (itr != mSource.begin() && (itr-1)->overlap(newent)) {
	source_iter jtr = itr - 1;
	jtr->merge(newent);
	if (jtr->overlap(*itr)) {
	    jtr->merge(*itr);
	    mSource.erase(itr);
	}
    } else if (itr != mSource.end() && itr->overlap(newent)) {
	itr->merge(newent);
    } else {
	//------------------------------  Save recalculating the lower_bound
	static_cast<source_vect::base_type&>(mSource).insert(itr, newent);
    }
}

//======================================  Clear channel entry
void
chan_entry::clear(void) {
    mName.clear();
    mSource.clear();
}

//======================================  Search source list value
bool 
chan_entry::is_source(const std::string& src, gps_type t) const {
    source_count N = mSource.size();
    for (source_count i=0; i<N; ++i) {
	if (mSource[i].name() == src && mSource[i].test_gps(t)) return true;
    }
    return false;
}

//======================================  Merge source entries
//
//   The trivial solution (svn 1118) requires three steps: merge, sort 
//   and compress. This can be sped up (especially in cacses of adding 
//   overlapping segments) by making a first pass to count the changes 
//   and determine whether a second pass is necessary.
//
void 
chan_entry::merge(const chan_entry& ce) {
    //----------------------------------  Trivial solution if no sources.
    if (ce.mSource.empty()) {
	return;
    } else if (mSource.empty()) {
	mSource = ce.mSource;
	return;
    }

    //----------------------------------  Automatiion
    //  1 push first operand
    //  2 push second operand
    //  3 merge first operand
    //  4 merge second operand
    std::vector<int> autom;
    autom.reserve(16);

    //----------------------------------  Get vector limits.
    const_source_iter pi = mSource.begin();
    const_source_iter ei = mSource.end();
    const_source_iter pj = ce.mSource.begin();
    const_source_iter ej = ce.mSource.end();

    //----------------------------------  Count added entries
    int nOut = 0;
    bool rebuild = false;
    source_entry se;
    if (*pj < *pi) {
	se = *pj++;
	nOut++;
	rebuild = true;
	autom.push_back(2);
    } else {
	se = *pi++;
	autom.push_back(1);	
    }
    while (pi != ei && pj != ej) {
	if (*pj < *pi) {
	    if (se.overlap(*pj)) {
		se.merge(*pj++);
		autom.push_back(4);
	    } else {
		se = *pj++;
		autom.push_back(2);
		++nOut;
		rebuild = rebuild || (nOut > 1);
	    }
	} else {
	    if (se.overlap(*pi)) {
		se.merge(*pi++);
		autom.push_back(3);
		--nOut;
	    } else {
		se = *pi++;
		autom.push_back(1);
	    }
	}
    }
    //----------------------------------  Process remaining new list.
    if (pj != ej) {
	if (se.overlap(*pj)) {
	    se.merge(*pj++);
	    autom.push_back(4);
	}
	while (pj != ej) {
	    nOut++;
	    ++pj;
	    autom.push_back(2);
	}
    }

    //----------------------------------  Process remaining list.
    if (pi != ei) {
	if (se.overlap(*pi)) {
	    se.merge(*pi++);
	    --nOut;
	    autom.push_back(3);
	}
	while (pi != ei) {
	    ++pi;
	    autom.push_back(1);
	}
    }

    //----------------------------------  Test for merge in place
    pi = mSource.begin();
    pj = ce.mSource.begin();
    size_t  newSize = mSource.size() + nOut;
    if (rebuild || newSize > mSource.capacity()) {

	//------------------------------  Construct output vector
	source_vect::base_type sv;
	sv.reserve(newSize);
	size_t N = autom.size();
	for (size_t i=0; i<N; ++i) {
	    switch (autom[i]) {
	    case 1:
		sv.push_back(*pi++);
		break;
	    case 2:
		sv.push_back(*pj++);
		break;
	    case 3:
		sv.back().merge(*pi++);
		break;
	    case 4:
		sv.back().merge(*pj++);
		break;
	    }
	}
	mSource.swap(sv);
    }

    //----------------------------------  Merge in place
    else {
	if (nOut > 0) mSource.resize(newSize);
	source_iter pk = mSource.begin();
	source_iter pi = pk;
	if (autom[0] != 1) 
	    throw logic_error("channel_entry::merge doesn't work in-place");
	++pi;
	size_t N = autom.size();
	for (size_t i=1; i<N; ++i) {
	    switch (autom[i]) {
	    case 1:
		++pk;
		if (pi != pk) pk->swap(*pi);
		++pi;
		break;
	    case 2:
		*(++pk) = *pj++;
		break;
	    case 3:
		pk->merge(*pi++);
		break;
	    case 4:
		pk->merge(*pj++);
		break;
	    }
	}
	if (nOut < 0) mSource.resize(newSize);
    }
}

//======================================  Memory usage
index_type 
chan_entry::mem_use(void) const {
    index_type nBytes = sizeof(chan_entry) + string_size(name());
    index_type N = mSource.size();
    nBytes += mSource.capacity() * sizeof(source_entry);
    for (index_type i=0; i<N; ++i) {
	nBytes += string_size(mSource[i].name());
    }
    return nBytes;
}

//======================================  Swap two channel entries
void
chan_entry::swap(chan_entry& ce) {
    std::swap<hash_type>(mHash,   ce.mHash);
    std::swap<chan_type>(mType,   ce.mType);
    std::swap<long>     (mAccess, ce.mAccess);
    std::swap<data_type>(mDType,  ce.mDType);
    std::swap<double>   (mRate,   ce.mRate);
    mName.swap(ce.mName);
    mSource.swap(ce.mSource);
}

//======================================  Set the access level.
void
chan_entry::set_access(long lvl) {
    mAccess = lvl;
}

//======================================  Write a channel description
std::ostream& 
chan_entry::write(ostream& out, bool times) const {
    string tmp = mName;
    while (tmp.find(' ') != string::npos) tmp[tmp.find(' ')] = '~';
    out << tmp << " " << mType.str() << " " << mAccess << " " 
	<< mDType.str() << " " << mRate;
    index_type N = mSource.size();
    for (index_type i=0; i<N; ++i) {
	out << " " << mSource[i].name();
	if (times) out << " " << mSource[i].refEpoch();
    }
    out << endl;
    return out;
}



//======================================  Print chan_entry contents
std::ostream& 
sends::operator<<(std::ostream& out, const chan_entry& ent) {
    return out << ent.name() << "%" 
	       << ent.type().str();
}

//======================================  Print chan_entry contents
inline bool
operator<(const chan_list::index_node& a, const chan_list::index_node& b) {
    return a.first < b.first;
}

//======================================  Channel entry constructor
chan_list::chan_list(unsigned long ndef) 
    : list_available(false), index_valid(true)
{
    mChanList.reserve(ndef);
}

//======================================  Add a channel to the list.
chan_index
chan_list::add_channel(const chan_entry& ent) {
    chan_index inx = mChanList.size();
    mChanList.push_back(ent);
    mChanIndex.push_back(index_node(ent.hash(), inx));
    index_valid=false;
    return inx;
}

//======================================  Build the channel hash index.
void
chan_list::build_index(void) {
    size_t N = mChanList.size();
    if (index_valid && mChanIndex.size() == N) return;
    mChanIndex.clear();
    mChanIndex.reserve(N);
    for (size_t i=0; i<N; ++i) {
	hash_type key = mChanList[i].hash();
	mChanIndex.push_back(index_node(key, i));
    }
    mChanIndex.sort();
    index_valid = true;
}

//======================================  Remove all entries from the list.
void 
chan_list::clear(void) {
    list_available = false;
    mChanList.clear();
    mChanIndex.clear();
    index_valid = true;
}

#define DEBUG_COMPRESS 1

//======================================  Static function to merge duplicates.
static bool
merge_dup(chan_entry& a, chan_entry& b) {
    if (a == b) {
	a.merge(b);
	b.clear();
	return true;
    }
    return false;
}

//======================================  Build the channel hash index.
void
chan_list::compress(bool shrink) {
    iterate_isonym(merge_dup, shrink);
}

//===================================  Print all entries from the list.
std::ostream& 
chan_list::dump(std::ostream& out, bool times) const {
    out << "# Channel list: " << endl;
    out << "# Name  type  access d-type rate [Sources]" << endl;
    chan_index N=size();
    for (chan_index i=0; i<N; ++i) {
	mChanList[i].write(out, times);
    }
    return out;
}

//======================================  Allow access to the channel list
void 
chan_list::enable(void) {
    compress();
    list_available = true;
}

//======================================  Find a matching entry
chan_index 
chan_list::find(const chan_entry& ent) const {
    index_type lb = lower_bound(ent.name());
    index_type ub = upper_bound(ent.name());
    for (index_type inx=lb; inx<ub; ++inx) {
	chan_index chanId = mChanIndex[inx].second;
	if (mChanList[chanId] == ent) return chanId;
    }
    throw runtime_error(string("Unknown or unavailable channel: ")+ent.name());
}

//======================================  Find the specified entry
chan_index 
chan_list::find_best(const string& name, chan_type ctype,
		     double rate, gps_type gps) const {
    index_type lb = lower_bound(name);
    index_type ub = upper_bound(name);
    for (index_type inx=lb; inx<ub; ++inx) {
	chan_index chanId = mChanIndex[inx].second;
	if (chanId >= size()) continue;
	const chan_entry& ce = mChanList[chanId];
	if (ctype != chan_type::kUnknown && ctype != ce.type()) continue;
	if (rate != 0 && rate > ce.rate()) continue;
	if (gps  != 0) {
	    if (ce.type() == chan_type::kOnline) continue;
	    if (!ce.available(gps)) continue;
	}
	if (ce.name() != name) continue;
	return chanId;
    }
    throw runtime_error(string("Unknown or unavailable channel: ") + name);
}

//======================================  Iterate over pairs of isonym entries
bool 
chan_list::iterate_isonym(combine& algo, bool shrink) {
    bool rc = false;
    chan_entry invChan("", chan_type::kUnknown);
    if (!index_valid) build_index();
    size_t N = mChanList.size();

    //----------------------------------  merge and tag duplicates
    size_t nComp = 0;
    for (size_t in1=0; in1 < N; in1++) {
	chan_index inx1 = mChanIndex[in1].second;
	if (inx1 != N) {
	    chan_entry& ref1 = mChanList[inx1];
	    hash_type   key1 = mChanIndex[in1].first;
	    for (size_t in2=in1+1; in2 < N; in2++) {
		if (mChanIndex[in2].first != key1) break;
		chan_index inx2 = mChanIndex[in2].second;
		if (inx2 != N) {
		    chan_entry& ref2 = mChanList[inx2];
		    //------------------  Combine the entries. Invalidate 
		    //                    deleted entry.
		    if (algo(ref1, ref2)) rc = true;
		    if (ref2.empty()) {
			mChanIndex[in2].second = N;
		    }
		    if (ref1.empty()) {
			mChanIndex[in1].second = N;
			break;
		    }
		}
	    }
	    if ( !ref1.empty() ) nComp++;
	}
    }

    //----------------------------------  Build up the compressed list.
    if (nComp != N) {
	chan_vect newvec;
	if (shrink) newvec.reserve(nComp);
	else        newvec.reserve(capacity());
	for (chan_index i=0; i<N; ++i) {
	    if (!mChanList[i].empty()) newvec.push_back(mChanList[i]);
	}
	mChanList.swap(newvec);

	//------------------------------  Rebuild the hash index
	index_valid = false;
	build_index();
    }
    return rc;
}

//======================================  Find the specified entry
chan_index 
chan_list::lower_bound(const string& name) const {
    hash_type key = chan_entry::hash_key(name);
    return mChanIndex.lower_bound(index_node(key, 0)) - mChanIndex.begin();
}

//======================================  Find the specified entry
chan_index 
chan_list::upper_bound(const string& name) const {
    hash_type key = chan_entry::hash_key(name);
    return mChanIndex.upper_bound(index_node(key, 0)) - mChanIndex.begin();
}

//======================================  Merge a channel list
void
chan_list::merge(const chan_vect& vec) {
    mChanList.insert(mChanList.end(), vec.begin(), vec.end());
    index_valid = false;
}

//======================================  Memory usage
index_type 
chan_list::mem_use(void) const {
    index_type nBytes = sizeof(chan_list);
    index_type N = mChanList.size();
    for (index_type i=0; i<N; ++i) {
	nBytes += mChanList[i].mem_use();
    }
    return nBytes;
}

//======================================  Read channels from a file
void 
chan_list::read(const std::string& file, const epoch& ep) {
    const string null_str;
    if (!mChanList.capacity()) mChanList.reserve(channel_list_default_size);

    ParseLine pl(file.c_str());
    if (!pl.isOpen()) {
	cout << "*** Error: Missing channel list file: " << file <<endl;
	return;
    }

    size_t chan_count = 0;

    chan_entry null_ce(null_str, chan_type::kUnknown, 0, data_type::kByte1, 0);
    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (nArg < 2) {
	    ostringstream msg;
	    msg << "chan_list::read: Insufficient arguments in "
	        << file << ": " << pl.getLineNumber();
	    throw runtime_error(msg.str());
	}
	string name = pl[0];
	chan_type type;
	type.set(pl[1]);

	//------------------------------  Get access level
	long access = 0;
	if (nArg > 2) access = pl.getInt(2);

	//------------------------------  Get data type
	data_type dtype(data_type::kReal4);
	if (nArg > 3 && *pl[3] != '-') {
	    dtype.set(pl[3]);
	}

	//------------------------------  Get data rate
	double rate = 0.0;
	if (nArg > 4 && *pl[4] != '-') {
	    rate = pl.getDouble(4); 
	}

	//------------------------------  Get first source, build channel entry
	int i0 = 5;
	string source;
	epoch epik(0);
	if (i0 < nArg) {
	    source = pl[i0++];
	    if (i0<nArg && isdigit(*pl[i0])) epik = epoch::parse(pl[i0++], 0);
	    else                             epik = ep;
	}
	chan_entry ce(name, type, access, dtype, rate, source, epik);

	//------------------------------  Get source list
	bool  nope = false;           //  No epoch recorded for this source.
	for (int i=i0; i<nArg; ++i) {
	    if (isdigit(*pl[i])) {
		ce.addSource(source, epoch::parse(pl[i], 0));
		nope = false;
	    } else {
		if (nope) ce.addSource(source, ep);
		source = pl[i];
		nope = true;
	    }
	}
	if (nope) ce.addSource(source, ep);

	//------------------------------  Put data into temporary
	size_t vect_size = mChanList.size();
	if (vect_size == mChanList.capacity()) {
	    vect_size *= 2;
	    mChanList.reserve(vect_size);
	}
	mChanList.push_back(null_ce);
	mChanList.back().swap(ce);
	chan_count++;
    }
    index_valid = false;
    cout << "chan_list: " << chan_count << " channels read from: " 
	 << file << endl;
}

//======================================  Reserve space in the channel list
void 
chan_list::reserve(chan_index nMax) {
    mChanList.reserve(nMax);
    mChanIndex.reserve(nMax);
}

//======================================  Reserve space in the channel list
void 
chan_list::sort(void) {
    index_valid = false;
    std::sort(mChanList.begin(), mChanList.end());
}
